library in_multibutton;

(* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this file except in compliance with 
 * the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the MultiButton Input Module for Lightning.
 *
 * The Initial Developer of the Original Code is
 * Tomsoft.
 * Portions created by Tomsoft are Copyright (C) 2011
 * Tomsoft. All Rights Reserved.
 *
 * Contributor(s):
 *   Tommy Sools <tom@sools.com>
 *
 * ***** END LICENSE BLOCK ***** *)

{$IMAGEBASE $410D0000}

uses
  SysUtils,
  Classes,
  Windows,
  Graphics,
  Controls,
  StdCtrls,
  ExtCtrls,
  common in '..\common.pas',
  unitkeyedit in '..\unitkeyedit.pas';

{$R *.RES}
{$R ..\RESOURCES\BUTTONS.RES}
{$R ..\RESOURCES\OVERRIDE.RES}

const
  channels = 8;

type
  TButtonRecord = record
    image: TImage;
    l: TLabel;
    pressed: Boolean;
    key: Longint;
    keyShift, keyControl, keyScrollLock: Boolean;
  end;
  PButtonRecord = ^TButtonRecord;
  TDetails = class(TObject)
  private
    bitmapOverrideDown, bitmapOverrideUp: TBitmap;
    bitmapButtonDown, bitmapButtonUp: TBitmap;
    pButtons: TPanel;
    buttons: array[1..channels] of TButtonRecord;
    switch, exclusive: Boolean;
    procedure depressAllOthers(index: Integer);
    procedure changePressed(index: Integer);
    procedure setPressed(index: Integer; pressed: Boolean);
    procedure setKey(button: Integer; key: Longint; keyShift: Boolean; keyControl: Boolean; keyScrollLock: Boolean);
    function getKey(button: Integer): Longint;
  public
    gbDetails: TGroupBox;
    listBoxButtons: TListBox;
    labelKey: TLabel;
    editKey: TKeyEdit;
    checkBoxScrollLock: TCheckBox;
    checkBoxShift: TCheckBox;
    checkBoxControl: TCheckBox;
    chSwitch, chExclusive: TCheckBox;
    pOverride: TPanel;
    iOverride: TImage;
    overridden: Boolean;
    procedure iOverrideMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure iButtonMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure iButtonMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    function getButtonListIndex(): Integer;
    procedure incButtonListIndex();
    procedure buttonListChange(Sender: TObject);
    procedure keyShiftControl(Sender: TObject; shift: Boolean; control: Boolean);
    procedure keyChange(Sender: TObject);
    procedure chSwitchClick(Sender: TObject);
    procedure chExclusiveClick(Sender: TObject);
  end;

procedure TDetails.depressAllOthers(index: Integer);
var
  i: Integer;
begin
  for i := 1 to channels do
    if i <> index then
      setPressed(i, False);
end;

procedure TDetails.changePressed(index: Integer);
begin
  setPressed(index, not(buttons[index].pressed));
end;

procedure TDetails.setPressed(index: Integer; pressed: Boolean);
begin
  if buttons[index].pressed <> pressed then
  begin
    if pressed and exclusive then
      depressAllOthers(index);
    buttons[index].pressed := pressed;
    if pressed then
    begin
      buttons[index].image.Picture.Assign(bitmapButtonDown);
      buttons[index].l.Top := buttons[index].l.Top + 1;
      buttons[index].l.Left := buttons[index].l.Left + 1;
    end
    else
    begin
      buttons[index].image.Picture.Assign(bitmapButtonUp);
      buttons[index].l.Top := buttons[index].l.Top - 1;
      buttons[index].l.Left := buttons[index].l.Left - 1;
    end;
  end;
end;

procedure TDetails.iOverrideMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  overridden := not(overridden);
  if overridden then
    iOverride.Picture.Assign(bitmapOverrideDown)
  else
    iOverride.Picture.Assign(bitmapOverrideUp);
end;

procedure TDetails.iButtonMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if switch then
    changePressed(TComponent(Sender).Tag)
  else
    setPressed(TComponent(Sender).Tag, True);
end;

procedure TDetails.iButtonMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if not(switch) then
    setPressed(TComponent(Sender).Tag, False);
end;

function TDetails.getButtonListIndex(): Integer;
begin
  if listBoxButtons.ItemIndex > -1 then
    Result := listBoxButtons.ItemIndex + 1
  else
    Result := 1;
end;

procedure TDetails.incButtonListIndex();
begin
  if listBoxButtons.ItemIndex >= listBoxButtons.Items.Count - 1 then
    listBoxButtons.ItemIndex := 0
  else
    listBoxButtons.ItemIndex := listBoxButtons.ItemIndex + 1;
  buttonListChange(nil);
end;

procedure TDetails.buttonListChange(Sender: TObject);
var
  buttonIndex: Integer;
begin
  buttonIndex := getButtonListIndex();
  checkBoxShift.OnClick      := nil;
  checkBoxControl.OnClick    := nil;
  checkBoxScrollLock.OnClick := nil;
  try
    checkBoxShift.Checked := buttons[buttonIndex].keyShift;
    checkBoxControl.Checked := buttons[buttonIndex].keyControl;
    checkBoxScrollLock.Checked := buttons[buttonIndex].keyScrollLock;
  finally
    checkBoxShift.OnClick      := keyChange;
    checkBoxControl.OnClick    := keyChange;
    checkBoxScrollLock.OnClick := keyChange;
  end;
  editKey.setVirtualKey(getKey(buttonIndex));
end;

procedure TDetails.keyShiftControl(Sender: TObject; shift: Boolean; control: Boolean);
begin
  checkBoxShift.OnClick   := nil;
  checkBoxControl.OnClick := nil;
  try
    checkBoxShift.Checked   := shift;
    checkBoxControl.Checked := control;
  finally
    checkBoxShift.OnClick   := keyChange;
    checkBoxControl.OnClick := keyChange;
  end;
end;

procedure TDetails.keyChange(Sender: TObject);
begin
  setKey(
    getButtonListIndex(),
    editKey.getVirtualKey(),
    checkBoxShift.Checked,
    checkBoxControl.Checked,
    checkBoxScrollLock.Checked
  );
  if (Sender = editKey) and (editKey.getVirtualKey() <> 0) then
    incButtonListIndex()
  else
    editKey.SetFocus();
end;

procedure TDetails.chSwitchClick(Sender: TObject);
begin
  switch := chSwitch.Checked;
  if switch then
    chExclusive.Enabled := True
  else
  begin
    chExclusive.Enabled := False;
    chExclusive.Checked := False;
  end;
end;

procedure TDetails.chExclusiveClick(Sender: TObject);
begin
  exclusive := chExclusive.Checked;
end;

procedure TDetails.setKey(button: Integer; key: Longint; keyShift: Boolean; keyControl: Boolean; keyScrollLock: Boolean);
var
  buf: string;
begin
  buttons[button].key           := key;
  buttons[button].keyShift      := keyShift;
  buttons[button].keyControl    := keyControl;
  buttons[button].keyScrollLock := keyScrollLock;
  buf := TKeyEdit.translateKey(key);
  // only show hotkey in module's button if it's 1 character long and
  // requires no modifiers (Shift, Control).
  if not(keyShift)
     and
     not(keyControl)
     and
     (Length(buf) = 1) then
  begin
    if keyScrollLock then
      // show hotkey in red if scroll lock must be pressed
      buttons[button].l.Font.Color := clRed
    else
      // show hotkey in yellow if scroll lock must be depressed
      buttons[button].l.Font.Color := clYellow;
    buttons[button].l.Caption := buf;
    buttons[button].l.Visible := True;
  end
  else
  begin
    buttons[button].l.Caption := '';
    buttons[button].l.Visible := False;
  end;
end;

function TDetails.getKey(button: Integer): Longint;
begin
  Result := buttons[button].key;
end;

function  init(name: PChar; nameLength: DWORD): DWORD; cdecl;
begin
  Result := return('IMultiButton|Controls', name, nameLength);
end;

procedure finish(); cdecl;
begin
  //
end;

function  create(controlHandle: THandle): PLightningModule; cdecl;
var
  myDetails: TDetails;
  i:         Integer;
begin
  New(Result);
  with Result^ do
  begin
    handle        := controlHandle;
    title         := 'MultiButton';
    settings      := True;
    width         := -1;
    height        := 80;
    inputTop      := -1;
    infoTop       := -1;
    outputTop     := 56;
    inputs        := 0;
    outputs       := channels;
    inputTooltip  := '';
    GetMem(outputTooltip, maxLength);
    StrCopy(outputTooltip, '');
    for i := 1 to channels do
      StrCat(outputTooltip, PChar('Button ' + IntToStr(i) + ' status|'));
    myDetails     := TDetails.Create;
    reserved      := nil;
    details       := myDetails;
  end;
  with myDetails do
  begin
    overridden := False;
    bitmapOverrideUp := TBitmap.Create();
    with bitmapOverrideUp do
    begin
      LoadFromResourceName(hInstance, 'overrideup');
    end;
    bitmapOverrideDown := TBitmap.Create();
    with bitmapOverrideDown do
    begin
      LoadFromResourceName(hInstance, 'overridedown');
    end;
    pOverride := TPanel.Create(nil);
    with pOverride do
    begin
      Top        := 74;
      Left       := 1;
      BevelOuter := bvNone;
    end;
    iOverride := TImage.Create(pOverride);
    with iOverride do
    begin
      Cursor      := crHandPoint;
      AutoSize    := True;
      Hint        := 'Manual override';
      ShowHint    := True;
      Picture.Bitmap.LoadFromResourceName(hInstance, 'overrideup');
      OnMouseDown := iOverrideMouseDown;
      Parent      := pOverride;
    end;
    with pOverride do
    begin
      Width        := iOverride.Width;
      Height       := iOverride.Height;
      ParentWindow := controlHandle;
    end;
    switch    := False;
    exclusive := False;
    bitmapButtonUp := TBitmap.Create();
    with bitmapButtonUp do
    begin
      LoadFromResourceName(hInstance, 'buttonup');
    end;
    bitmapButtonDown := TBitmap.Create();
    with bitmapButtonDown do
    begin
      LoadFromResourceName(hInstance, 'buttondown');
    end;
    pButtons := TPanel.Create(nil);
    with pButtons do
    begin
      Top        := 32;
      Left       := 8;
      BevelOuter := bvNone;
      Width      := channels * (16 + 8);
      Height     := 16;
    end;
    for i := 1 to channels do
    begin
      buttons[i].pressed := False;
      buttons[i].image := TImage.Create(pButtons);
      with buttons[i].image do
      begin
        Top         := 0;
        Left        := (i - 1) * (16 + 8);
        Tag         := i;
        Cursor      := crHandPoint;
        AutoSize    := True;
        Picture.Bitmap.LoadFromResourceName(hInstance, 'buttonup');
        OnMouseDown := iButtonMouseDown;
        OnMouseUp   := iButtonMouseUp;
        Parent      := pButtons;
      end;
      buttons[i].l := TLabel.Create(pButtons);
      with buttons[i].l do
      begin
        Transparent := True;
        Cursor      := crHandPoint;
        AutoSize    := False;
        Top         := 0;
        Left        := (i - 1) * (16 + 8);
        Width       := buttons[i].image.Width;
        Height      := buttons[i].image.Height;
        Tag         := i;
        Alignment   := taCenter;
        Layout      := tlCenter;
        with Font do
        begin
          Name  := 'Arial';
          Size  := 7;
          Color := clYellow;
        end;
        Caption     := '';
        OnMouseDown := buttons[i].image.OnMouseDown;
        OnMouseUp   := buttons[i].image.OnMouseUp;
        Visible     := False;
        Parent      := pButtons;
      end;
    end;
    pButtons.ParentWindow := controlHandle;
  end;
end;

procedure destroy(lightningModule: PLightningModule); cdecl;
var
  i: Integer;
begin
  with TDetails(lightningModule^.details) do
  begin
    bitmapOverrideUp.Free();
    bitmapOverrideDown.Free();
    iOverride.Free();
    pOverride.Free();
    bitmapButtonUp.Free();
    bitmapButtonDown.Free();
    for i := 1 to channels do
    begin
      buttons[i].image.Free();
      buttons[i].l.Free();
    end;
    pButtons.Free();
    Free();
  end;
  FreeMem(lightningModule^.outputTooltip);
  Dispose(lightningModule);
end;

function  action(lightningModule: PLightningModule; input, output: Integer; inputConnected, outputConnected: PChar; var bridge: TSpectrum): Boolean; cdecl;
begin
  Result := True;
  if TDetails(lightningModule.details).buttons[output].pressed then
    createSpectrum(bridge, levelMax)
  else
    createSpectrum(bridge, levelMin);
end;

procedure show(lightningModule: PLightningModule; handle: THandle); cdecl;
begin
  with TDetails(lightningModule.details) do
  begin
    gbDetails := TGroupBox.Create(nil);
    with gbDetails do
    begin
      Top          := detailsTop;
      Left         := detailsLeft;
      Width        := detailsWidth;
      Height       := detailsHeight;
      Ctl3D        := True;
      Color        := clBtnFace;
      Caption      := 'MultiButton Trigger Settings';
    end;
    listBoxButtons := TListBox.Create(gbDetails);
    with listBoxButtons do
    begin
      Top     := 20;
      Left    := 8;
      Width   := 40;
      Height  := 108;
      OnClick := buttonListChange;
      Parent  := gbDetails;
    end;
    checkBoxShift := TCheckBox.Create(gbDetails);
    with checkBoxShift do
    begin
      Top     := 20;
      Left    := 76;
      Width   := 60;
      Caption := 'Shift +';
      OnClick := keyChange;
      Parent  := gbDetails;
    end;
    checkBoxControl := TCheckBox.Create(gbDetails);
    with checkBoxControl do
    begin
      Top     := 40;
      Left    := 76;
      Width   := 60;
      Caption := 'Ctrl +';
      OnClick := keyChange;
      Parent  := gbDetails;
    end;
    labelKey := TLabel.Create(gbDetails);
    with labelKey do
    begin
      AutoSize := True;
      Top      := 20;
      Left     := 136;
      Caption  := 'Key:';
      Parent   := gbDetails;
    end;
    editKey := TKeyEdit.Create(gbDetails);
    with editKey do
    begin
      Top            := 40;
      Left           := 136;
      Width          := 100;
      ReadOnly       := True;
      Tag            := Longint(TDetails(lightningModule^.details));
      OnShiftControl := keyShiftControl;
      OnChange       := keyChange;
      Parent         := gbDetails;
    end;
    checkBoxScrollLock := TCheckBox.Create(gbDetails);
    with checkBoxScrollLock do
    begin
      Top       := 40;
      Left      := 246;
      Width     := 130;
      Caption   := '+ Scroll Lock enabled';
      Alignment := taLeftJustify;
      OnClick   := keyChange;
      Parent    := gbDetails;
    end;
    chSwitch := TCheckBox.Create(gbDetails);
    with chSwitch do
    begin
      Top        := 20;
      Left       := 468;
      Width      := 300;
      Caption    := 'Act as switches';
      Checked    := switch;
      OnClick    := chSwitchClick;
      Parent     := gbDetails;
    end;
    chExclusive := TCheckBox.Create(gbDetails);
    with chExclusive do
    begin
      Top        := 40;
      Left       := 468;
      Width      := 300;
      Caption    := 'Mutually exclusive';
      Checked    := exclusive;
      Enabled    := switch;
      OnClick    := chExclusiveClick;
      Parent     := gbDetails;
    end;
    gbDetails.ParentWindow := handle;
    with listBoxButtons do
    begin
      Items.Add('1');
      Items.Add('2');
      Items.Add('3');
      Items.Add('4');
      Items.Add('5');
      Items.Add('6');
      Items.Add('7');
      Items.Add('8');
      ItemIndex := 0;
      buttonListChange(nil);
    end;
  end;
end;

procedure hide(lightningModule: PLightningModule); cdecl;
begin
  with TDetails(lightningModule.details) do
  begin
    listBoxButtons.Free();
    checkBoxShift.Free();
    checkBoxControl.Free();
    labelKey.Free();
    editKey.Free();
    checkBoxScrollLock.Free();
    chExclusive.Free();
    chSwitch.Free();
    gbDetails.Free();
  end;
end;

procedure open(lightningModule: PLightningModule; settings: PChar); cdecl;
var
  details: TDetails;
  bString, flags: string;
  i: Integer;
  key: Longint;
begin
  details           := TDetails(lightningModule^.details);
  bString           := settings;
  details.switch    := parse(bString, ',') <> '';
  details.exclusive := parse(bString, ',') <> '';
  for i := 1 to channels do
    details.setPressed(i, parse(bString, ',') <> '');
  for i := 1 to channels do
  begin
    key := StrToIntDef(parse(bString, ','), 0);
    flags := parse(bString, ',');
    details.setKey(
      i,
      key,
      Pos('S', flags) > 0,
      Pos('C', flags) > 0,
      Pos('*', flags) > 0
    );
  end;
end;

function  save(lightningModule: PLightningModule; settings: PChar; settingsLength: DWORD): DWORD; cdecl;
var
  details: TDetails;
  bString, flags: string;
  i: Integer;
begin
  details := TDetails(lightningModule^.details);
  bString := '';
  if details.switch then
    bString := bString + '1';
  bString := bString + ',';
  if details.exclusive then
    bString := bString + '1';
  bString := bString + ',';
  for i := 1 to channels do
  begin
    if details.buttons[i].pressed then
      bString := bString + '1';
    bString := bString + ',';
  end;
  for i := 1 to channels do
  begin
    flags := '';
    if details.buttons[i].keyShift then
      flags := flags + 'S';
    if details.buttons[i].keyControl then
      flags := flags + 'C';
    if details.buttons[i].keyScrollLock then
      flags := flags + '*';
    bString := bString + IntToStr(details.getKey(i)) + ',' + flags + ',';
  end;
  Result := return(bString, settings, settingsLength);
end;

procedure openscene(lightningModule: PLightningModule; settings: PChar); cdecl;
var
  bString: string;
  i: Integer;
begin
  bString := settings;
  with TDetails(lightningModule^.details) do
  begin
    if not(overridden) then
    begin
      for i := 1 to channels do
        setPressed(i, parse(bString, ',') <> '');
    end;
  end;
end;

function  savescene(lightningModule: PLightningModule; settings: PChar; settingsLength: DWORD): DWORD; cdecl;
var
  bString: string;
  i: Integer;
begin
  with TDetails(lightningModule^.details) do
  begin
    bString := '';
    for i := 1 to channels do
    begin
      if buttons[i].pressed then
        bString := bString + '1';
      bString := bString + ',';
    end;
    Result := return(bString + ',', settings, settingsLength);
  end;
end;

procedure keyboard(lightningModule: PLightningModule; nVirtKey: LongInt; lKeyData: Cardinal); cdecl;
var
  details: TDetails;
  i: Integer;
begin
  details := TDetails(lightningModule^.details);
  for i := 1 to channels do
  begin
    if (nVirtKey = details.getKey(i)) then
    begin
      // check if this is a key up to down transition
      if (details.buttons[i].keyShift = (HIBYTE(GetKeyState(VK_SHIFT)) > 0))
         and
         (details.buttons[i].keyControl = (HIBYTE(GetKeyState(VK_CONTROL)) > 0))
         and
         (details.buttons[i].keyScrollLock = (LOBYTE(GetKeyState(VK_SCROLL)) > 0))
         and
         ((lKeyData and $40000000) = 0)
         and
         ((lKeyData and $80000000) = 0) then
      begin
        if details.switch then
          details.changePressed(i)
        else
          details.setPressed(i, True);
      end
      // check if this is a key down to up transition
      else if (lKeyData and $80000000) > 0 then
      begin
        if not(details.switch) then
          details.setPressed(i, False);
      end;
    end;
  end;
end;

exports
  init, finish, destroy, create, action, show, hide, open, save, openscene, savescene, keyboard;

end.
